@@ -1,7 +1,8 @@
 from contextlib import contextmanager
 from typing import Generator, Optional
-from psycopg2.extensions import cursor
-from psycopg2.pool import ThreadedConnectionPool
+from psycopg_pool import ConnectionPool
+from psycopg import connect
+from psycopg import cursor
 from os import environ as env
 
 
@@ -27,7 +28,7 @@ class CoreDatabase:
         return {
             'host': env.get('DB_HOST', 'localhost'),
             'port': env.get('DB_PORT', 5432),
-            'database': env.get('DB_DATABASE', 'postgres'),
+            'dbname': env.get('DB_DATABASE', 'postgres'),
             'user': env.get('DB_USER', 'postgres'),
             'password': env.get('DB_PASSWORD', None)
         }
@@ -68,7 +69,7 @@ class CoreDatabase:
         cxn = {
             'host': env.get('DB_HOST', 'localhost'),
             'port': env.get('DB_PORT', 5432),
-            'database': env.get('DB_DATABASE', 'postgres'),
+            'dbname': env.get('DB_DATABASE', 'postgres'),
             'user': env.get('DB_USER', 'postgres'),
             'password': env.get('DB_PASSWORD', None)
         }
@@ -85,14 +86,16 @@ class CoreDatabase:
 
         return cls(cxn, **kwargs)
 
-    def __init__(self, cxn_config: dict, **kwargs):
+    def __init__(self, cxn_config: dict, pool_config: Optional[dict] = None, **kwargs):
         """
 
         Parameters
         ----------
         cxn_config : dict
-            connection parameters fed to Psycopg connection init (ie host, port, user, password, db, etc)
-        kwargs : any
+            connection parameters fed to Psycopg connection init (ie host, port, user, password, dbname, etc)
+            See https://www.postgresql.org/docs/current/libpq-connect.html#LIBPQ-PARAMKEYWORDS
+
+        kwargs : dict
             keyword argument added for super classing support for class methods
         """
         # parse connection parameters from ini file and merge
@@ -100,44 +103,20 @@ class CoreDatabase:
         self.db = None
         self._tables = None
         self.config = cxn_config
-        self.pool = ThreadedConnectionPool(1, 5, **self.config)
+        self.params = {'min_size': 2, 'open': True} | (pool_config if pool_config else {})
+        self.pool = ConnectionPool(kwargs=self.config, **self.params)
+        self.warnings = kwargs.get('warnings', True)
         self.__post_init__()
 
     def __post_init__(self):
         self.db = self.query('SELECT current_database()', as_tuple=True)[0]
 
-    @contextmanager
-    def cursor(self) -> Generator[cursor, None, None]:
-        """
-        Enable auto-magic cursor formation from pool, commit,
-        and return via context management.
-
-        Returns
-        -------
-        Generator
-            temporary cursor object
-        """
-
-        # Pull cxn from pool
-        cxn = self.pool.getconn()
-        cxn.autocommit = True
-        curs: cursor = cxn.cursor()
-
-        try:
-            # Provide
-            yield curs
-        finally:
-            # Cleanup on context end
-            cxn.commit()
-            curs.close()
-            cxn.close()
-            self.pool.putconn(cxn)
-
     def query(self,
               query: str,
               vals: tuple | None = None,
               as_tuple: bool | None = False,
-              no_resp: bool | None = False
+              no_resp: bool | None = False,
+              tidx: int | None = None,
               ) -> list:
         """
         Committed database query execution
@@ -146,15 +125,22 @@ class CoreDatabase:
         ----------
         query : str
             the command to be executed
+
         vals : tuple
             parameterized values to be included with query if applicable
             (default=None)
+
         as_tuple : bool
             modify return value to be a tuple list instead of dict
+
         no_resp : bool
             whether to attempt to fetch and parse execution output,
             ie from a select query (default=False)
 
+        tidx : int
+            allows tuple return to be preparsed to the given index
+            [commonly] tidx=0 will return values instead of a tuple singleton
+
         Returns
         -------
         list
@@ -162,16 +148,28 @@ class CoreDatabase:
         """
 
         fetched, cols = [], []
+        if tidx is not None and not as_tuple:
+            if self.warnings:
+                print('tidx option requires tuple return, enabling.')
+            as_tuple = True
 
         # execute with temp cursor
-        with self.cursor() as csr:
-            csr.execute(query, vals)
-            if not no_resp:
-                # grab columns for record formation
-                if not as_tuple:
-                    cols = tuple(x[0] for x in csr.description)
-                # retrieve values
-                fetched = csr.fetchall()
+        with self.pool.connection() as cxn:
+            with cxn.cursor() as csr:
+                csr: cursor.Cursor
+                csr.execute(query, vals)
+                try:
+                    if not no_resp:
+                        # grab columns for record formation
+                        if not as_tuple:
+                            cols = tuple(x[0] for x in csr.description)
+                        # retrieve values
+                        fetched = csr.fetchall()
+                except TypeError:
+                    if self.warnings:
+                        print('WARNING: no_resp not specified but query returned nothing.')
+                    cols = None
+                    fetched = None
 
         # cursor returned to pool before parsing, if required.
         if fetched:
@@ -180,6 +178,9 @@ class CoreDatabase:
             elif as_tuple:
                 return [x for x in fetched]
 
+        elif not no_resp:
+            return []
+
     @property
     def tables(self):
         if not self._tables:
@@ -250,7 +251,7 @@ class CoreDatabase:
         # a generator), its returning and iterable.
         return [x[0] for x in resp] if single else [x for x in resp]
 
-    def insert(self, table: str, data: dict | list, upsert_on: str | None = None) -> None:
+    def insert(self, table: str, data: dict | list, upsert_on: str | None = None, fill_na=False) -> None:
         """
         Fast parameterized insertion
 
@@ -263,7 +264,8 @@ class CoreDatabase:
             specified table fields.
         upsert_on : str
             unique column key for which collisions result in an update (default = None)
-
+        fill_na : bool
+            attempts to fill missing keys with None (default = False)
         Raises
         ------
         psycopg2.errors.UndefinedTable
@@ -280,10 +282,10 @@ class CoreDatabase:
         # force list
         data = [data] if isinstance(data, dict) else data
 
-        # extract keys from first - data uniformity is assumed, could
-        # be checked for with:
-        # kls = tuple(max(data, key=lambda x: len(data[x])).keys())
-        kls = tuple(data[0].keys())
+        # data uniformity is NOT assumed, finds insertion keys
+        kls = tuple(max(data, key=len).keys())
+
+        # kls = tuple(data[0].keys())
 
         # sql formatted table keys generated
         ks = f'({", ".join(kls)})'
@@ -295,7 +297,10 @@ class CoreDatabase:
         vks = ', '.join(kvs for _ in data)
 
         # grab all values and flatten to tuple
-        vs = tuple(z for x in data for z in x.values())
+        if fill_na:
+            vs = tuple(p.get(k) for p in data for k in kls)
+        else:
+            vs = tuple(p[k] for p in data for k in kls)
         # if previous data non-uniformity handling is implemented,
         # missing keys must instead be filled with:
         # vs = tuple(z for x in data for z in tuple(data[x].get(k, None) for k in kls))
